program MMC_M64I2C;
(*************************************************************)
(* a test program example which shows how to build           *)
(* a stand alone FAT16 MMC drive with the AVRco and a mega64 *)
(* The communication is build with a TWI Slave and Master    *)
(* This part is the TWI Slave which does the File accesses   *)
(* The Master counter part "MMC_M64i2cHost.pas" is included  *)
(* in the AVRco Demos directory in the "M64_MMC" folder      *)
(* Complete assembled and tested hardware can be ordered in  *)
(* the E-LAB internet shop                                   *)
(*************************************************************)

Device = mega64, VCC = 3.3;

Define_Fuses
  LockBits0      = [];
  FuseBits0      = [BODEN];
  FuseBits1      = [CKOPT, SPIEN, JTAGEN, OCDEN];
  FuseBits2      = [];
  ProgMode       = JTAG;

Import SysTick, TWIslave, FAT16;

From System Import LongWord, LongInt;

Define
  ProcClock      = 16000000;        {Hertz}
  SysTick        = 10;              {msec}
  StackSize      = $0050, iData;
  FrameSize      = $0100, iData;
  FAT16          = MMC_SPI, iData;
  F16_FileHandles = 2;
  F16_DirLevels  = 2;
  TWIbuffer      = 128, iData;
  TWIaddr        = 15;
  TWImode        = HandShake;

uses UFAT16;

Implementation

{$IDATA}

{--------------------------------------------------------------}
{ Type Declarations }

type
  tFileCmd    = (cmdSync, cmdGetState, cmdSetAddr, cmdRestart, cmdDiskFree,
                 cmdDiskSize, cmdFileAttr, cmdEraseAll, cmdCreateDir, cmdEraseDir,
                 cmdChangeDir, cmdPathExist, cmdCurDir, cmdFileExist, cmdFileSize,
                 cmdFileRename, cmdFileCopy, cmdFileDelete, cmdListFiles,
                 cmdFileAssign, cmdFileOpen, cmdCheckHandle, cmdFilePos, cmdFileSeek,
                 cmdBlockRead, cmdBlockWrite, cmdFileClose, cmdFileCreate,
                 cmdRandWrite, cmdEndOfFile, cmdFileSizeH, cmdRawAccess);

  tDiskState  = (dsNoDisk, dsWriteProt, dsDiskErr, dsDiskOk);

  tMediaState = (msUnKnown, msDiskChanged, msInit);

{--------------------------------------------------------------}
{ Const Declarations }
{$IDATA}
const
  ACK       : char  = #6;
  LF        : char  = #10;
  CR        : char  = #13;
  NAK       : char  = #21;
  ESC       : char  = #27;

{--------------------------------------------------------------}
{ Var Declarations }
{$PDATA}
var
  CardDet[@PinD, 3]     : bit;
  LED[@PortE, 7]       : bit;

{$IDATA}
var
  test: Longword;
  DiskState      : tDiskState;
  MediaState     : tMediaState;
  DiskError      : tDiskError;
  FileCmd        : tFileCmd;
  FPath, FPath1  : TPathStr;
  FName, FName1  : TFileName;
  fAttr          : tFAttr;
  fTime, fDate   : Word;
  SR             : TSearchRec;
  Fhw            : word;
  Fh[@Fhw]       : File;
  HandleOk       : boolean;
  FileMode       : TFileAccess;
  ww             : word;
  bb             : byte;
  lw             : longword;
  CommErr        : byte;
  SectBuff       : array[0..511] of byte;
  pb             : pointer;

  // TWI buffer access
  RxPtr          : pointer;
  TxPtr          : pointer;
  dd: tdiskerror;
{--------------------------------------------------------------}
{ functions }
procedure PortInit;
begin
  DDRA:= $00;                            // all input
  PortA:= $FF;                           // all pullups on
  DDRB:= DDRB or $01;                    // MMC select as output
  PortB:= $FF;                           // all pullups on
  DDRC:= $00;                            // all input
  PortC:= $FF;                           // all pullups on
  DDRD:= DDRD and %00000011;             // change all bits to input except TWI bits

  PortD:= $FF;                           // all pullups on
  DDRE:= $80;                            // all input, LED output
  PortE:= $FF;                           // all pullups on
  DDRF:= $00;                            // all input
  PortF:= $FF;                           // all pullups on
  DDRG:= $00;                            // all input
  PortG:= $FF;                           // all pullups on
end;

{$DEVICE}
function RxChar : char;
begin
  return(char(RxPtr^++));
end;

function RxHexByte : byte;
var
  st     : string[2];
  b      : byte;
begin
  Read(RxChar, st, 2);
  b:= HexToInt(st);
  if RunErr then
    inc(CommErr);
  endif;
  ClearRunErr;
  return(b);
end;

function RxHexWord : word;
var
  st     : string[4];
  w      : word;
begin
  Read(RxChar, st, 4);
  w:= HexToInt(st);
  if RunErr then
    inc(CommErr);
  endif;
  ClearRunErr;
  return(w);
end;

function RxHexInt : integer;
var
  st     : string[4];
  i      : integer;
begin
  Read(RxChar, st, 4);
  i:= HexToInt(st);
  if RunErr then
    inc(CommErr);
  endif;
  ClearRunErr;
  return(i);
end;

function RxHexLw : longword;
var
  st     : string[8];
  lw     : longword;
begin
  Read(RxChar, st, 8);
  lw:= HexToInt(st);
  if RunErr then
    inc(CommErr);
  endif;
  ClearRunErr;
  return(lw);
end;

function RxHexLI : longint;
var
  st     : string[8];
  li     : longint;
begin
  Read(RxChar, st, 8);
  li:= HexToInt(st);
  if RunErr then
    inc(CommErr);
  endif;
  ClearRunErr;
  return(li);
end;

procedure RxFilePath;
begin
  Read(RxChar, fPath, #0);
  ClearRunErr;
end;

procedure RxFileName;
begin
  Read(RxChar, FName, #0);
  ClearRunErr;
end;

function RxFileHandle : boolean;
begin
  Fhw:= RxHexWord;
  FileMode:= F16_CheckHandle(Fh);
  if FileMode = faNone then
    HandleOk:= false;
    return(false);
  else
    HandleOk:= true;
    return(true);
  endif;
end;

procedure RxFileAttr;
begin
  fAttr:= tFAttr(RxHexByte);
end;

procedure RxFileTime;
begin
  fTime:= RxHexWord;
end;

procedure RxFileDate;
begin
  fDate:= RxHexWord;
end;

procedure RxBlock(count : word);
var
  pt      : pointer;
begin
  pt:= @SectBuff;
  while count > 0 do
    pt^++:= byte(RxChar);
    dec(count);
  endwhile;
end;


{$DEVICE}
procedure TxChar(ch : char);
begin
  TxPtr^++:= byte(ch);
end;

procedure TxBlock(count : word);
var
  pt      : pointer;
begin
  pt:= @SectBuff;
  while count > 0 do
    TxChar(char(pt^++));
    dec(count);
  endwhile;
end;

procedure SendResponse;
begin
  TWIsetRDY(true);
  repeat
  until TWIgetTxStat;
  TxPtr:= @TWItxBuffer;
  TWIsetRDY(true);
end;


Function RestartFileSys : boolean;
begin
// DiskState  = (dsNoDisk, dsWriteProt, dsDiskErr, dsDiskOk);
// MediaState = (msUnKnown, msDiskChanged, msInit);
  if Not CardDet then
    return(false);
  endif;
  DiskState:= dsDiskErr;
  if not F16_DiskInit then
    dd := F16_Getdiskerror;
    return(false);
  endif;
  if not F16_DiskReset then
    return(false);
  endif;
  DiskState:= dsDiskOk;
  MediaState:= msInit;
  return(true);
end;


{--------------------------------------------------------------}
{ Main Program }
{$IDATA}

begin
  PortInit;
  Twisetslaveaddr(15);
  EnableInts;
  Loop
    Repeat
      LED:= Not CardDet;
    Until TWIgetrxstat;
    FileCmd:= tFileCmd(TWIgetCMD);
    RxPtr:= @TWIrxBuffer;
    if FileCmd <> cmdGetState then
      // clear previous errors
      CommErr:= 0;
    endif;

    case FileCmd of
      cmdSync       : // do nothing
                      F16_GetDiskError;
                      DiskError:= deNone;
                      TxPtr:= @TWItxBuffer;
                      TWIsetRDY(true);
                    |
      cmdGetState   : // return comm, drive and file states
                      DiskError:= F16_GetDiskError;
                      write(TxChar, cmdGetState + DiskState + MediaState + DiskError + CommErr + F16_GetUsedHandles);
                     CommErr:= 0;
                     SendResponse;
                    |
      cmdSetAddr    : // set Slave address
                      TxChar(char(cmdSetAddr));
                      bb:= RxHexByte;
                      if CommErr <> 0 then
                        TxChar(NAK);
                      else
                        TWIsetSlaveAddr(bb);
                      endif;
                      SendResponse;
                    |
      cmdRestart    : // Restart FileSys
                      TxChar(char(cmdRestart));
                      if RestartFileSys then
                        TxChar(ACK);
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdDiskFree   : // Send DiskFree
                      TxChar(char(cmdDiskFree));
                      if MediaState = msInit then
                        TxChar(ACK);
                     Test:=  F16_GetDiskfree;
                     Write(TxChar, LongToHex(Test));
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdDiskSize   : // SendDiskSize
                      TxChar(char(cmdDiskSize));
                      if MediaState = msInit then
                        TxChar(ACK);
                        Write(TxChar, LongToHex(F16_GetDiskSize));
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileAttr   : // read or write a file attribute
                      TxChar(char(cmdFileAttr));
                      case RxChar of
                        'G' : // Get File attribute
                              TxChar('G');
                              RxFilePath;
                              RxFileName;
                              if (MediaState = msInit) and (CommErr = 0) then
                                if F16_FileGetAttr(fPath, FName, fAttr) then
                                  TxChar(ACK);
                                  Write(TxChar, ByteToHex(fAttr));
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                        'S' : // Set File attribute
                              TxChar('S');
                              RxFilePath;
                              RxFileName;
                              RxFileAttr;
                              if (MediaState = msInit) and (CommErr = 0) then
                                if DiskState  = dsWriteProt then
                                  TxChar(NAK);
                                elsif F16_FileSetAttr(fPath, FName, fAttr) then
                                  TxChar(ACK);
                                  Write(TxChar, ByteToHex(fAttr));
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                      else
                        TxChar('?');
                      endcase;
                      SendResponse;
                    |
      cmdEraseAll   : // Remove all files
                      TxChar(char(cmdEraseAll));
                      if MediaState = msInit then
                        if DiskState  = dsWriteProt then
                          TxChar(NAK);
                        elsif F16_DiskFormat then
                          F16_DiskReset;
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdCreateDir  : // create a new directory
                      TxChar(char(cmdCreateDir));
                      RxFilePath;
                      RxFileName;
                      RxFileTime;
                      RxFileDate;
                      if (MediaState = msInit) and (CommErr = 0) then
                        if DiskState  = dsWriteProt then
                          TxChar(NAK);
                        elsif F16_CreateDir(fPath, fName, fTime, fDate) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdEraseDir   : // delete a directory
                      TxChar(char(cmdEraseDir));
                      RxFilePath;
                      RxFileName;
                      if MediaState = msInit then
                        if DiskState  = dsWriteProt then
                          TxChar(NAK);
                        elsif F16_RemoveDir(fPath, FName) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdChangeDir  : // change the default directory
                      TxChar(char(cmdChangeDir));
                      RxFilePath;
                      if MediaState = msInit then
                        if F16_ChangeDir(fPath) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdPathExist  : // check a path
                      TxChar(char(cmdPathExist));
                      RxFilePath;
                      if MediaState = msInit then
                        if F16_PathExist(fPath) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdCurDir     : // get the current directory
                      TxChar(char(cmdCurDir));
                      if MediaState = msInit then
                        fPath:= F16_GetCurDir;
                        TxChar(ACK);
                        Write(TxChar, fPath+#0);
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileExist  : // check a file existance
                      TxChar(char(cmdFileExist));
                      RxFilePath;
                      RxFileName;
                      RxFileAttr;
                      if (MediaState = msInit) and (CommErr = 0) then
                        if F16_FileExist(fPath, fName, fAttr) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileSize   : // return the byte count of a file
                      TxChar(char(cmdFileSize));
                      RxFilePath;
                      RxFileName;
                      if MediaState = msInit then
                        if F16_FileSize(fPath, fName, lw) then
                          TxChar(ACK);
                          Write(TxChar, LongToHex(lw));
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileRename : // rename a file
                      TxChar(char(cmdFileRename));
                      RxFilePath;
                      RxFileName;
                      FName1:= FName;
                      RxFileName;
                      if MediaState = msInit then
                        if DiskState  = dsWriteProt then
                          TxChar(NAK);
                        elsif F16_FileRename(fPath, FName1, FName) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileCopy   : // copy a file
                      TxChar(char(cmdFileCopy));
                      RxFilePath;
                      FPath1:= FPath;
                      RxFileName;
                      FName1:= FName;
                      RxFilePath;
                      RxFileName;
                      if MediaState = msInit then
                        if DiskState  = dsWriteProt then
                          TxChar(NAK);
                        elsif F16_FileCopy(FPath1, FName1, FPath, FName) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileDelete : // delete a file
                      TxChar(char(cmdFileDelete));
                      RxFilePath;
                      RxFileName;
                      if MediaState = msInit then
                        if DiskState  = dsWriteProt then
                          TxChar(NAK);
                        elsif F16_FileDelete(fPath, fName) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdListFiles  : // list files
                      TxChar(char(cmdListFiles));
                      RxFilePath;
                      RxFileName;
                      RxFileAttr;
                      if (MediaState = msInit) and (CommErr = 0) then
                        if F16_FindFirst(fPath, fName, fAttr, sr) then
                          TxChar(ACK);
                          Write(TxChar, sr.Name + #0);
                          Write(TxChar, ByteToHex(sr.Dir.FAttr));
                          SendResponse;
                          while RxChar = CR do
                            if F16_FindNext(sr) then
                              TxChar(ACK);
                              Write(TxChar, sr.Name + #0);
                              Write(TxChar, ByteToHex(sr.Dir.FAttr));
                              SendResponse;
                            else
                              break;
                            endif;
                          endwhile;
                          TxChar(NAK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileAssign : // assign a file and get the handle
                      TxChar(char(cmdFileAssign));
                      RxFilePath;
                      RxFileName;
                      if MediaState = msInit then
                        if F16_FileAssign(fh, fPath, fName) then
                          TxChar(ACK);
                          Write(TxChar, IntToHex(word(fh)));
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdCheckHandle: // check a handle
                      TxChar(char(cmdCheckHandle));
                      RxFileHandle;
                      if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                        TxChar(ACK);
                        TxChar(char(FileMode));
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileOpen   : // open a file for read or write access
                      TxChar(char(cmdFileOpen));
                      case RxChar of
                        'R' : // open a file for reading
                              TxChar('R');
                              RxFileHandle;
                              if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                                if F16_FileReset(fh) then
                                  TxChar(ACK);
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                        'W' : // open a file for writing
                              TxChar('W');
                              RxFileHandle;
                              RxFileAttr;
                              RxFileTime;
                              RxFileDate;
                              if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                                if DiskState  = dsWriteProt then
                                  TxChar(NAK);
                                elsif F16_FileRewrite(fh, fAttr, fTime, fDate) then
                                  TxChar(ACK);
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                        'A' : //  open a file for write appends
                              TxChar('A');
                              RxFileHandle;
                              if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                                if DiskState  = dsWriteProt then
                                  TxChar(NAK);
                                elsif F16_FileAppend(fh) then
                                  TxChar(ACK);
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                        'M' : //  open a file for random writing
                              TxChar('M');
                              RxFileHandle;
                              if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                                if DiskState  = dsWriteProt then
                                  TxChar(NAK);
                                elsif F16_RandomWrite(fh) then
                                  TxChar(ACK);
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                      else
                        TxChar('?');
                      endcase;
                      SendResponse;
                    |
      cmdFilePos    : // get the current pointer into a file
                      TxChar(char(cmdFilePos));
                      RxFileHandle;
                      if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                        lw:= F16_FilePos(fh);
                        Write(TxChar, ACK + LongToHex(lw));
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileSeek   : // set the file pointer to a position
                      TxChar(char(cmdFileSeek));
                      RxFileHandle;
                      lw:= RxHexLw;
                      if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                        lw:= F16_FileSeek(fh, lw);
                        Write(TxChar, ACK + LongToHex(lw));
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdBlockRead  : // read a block of data out of a file
                      TxChar(char(cmdBlockRead));
                      RxFileHandle;
                      ww:= RxHexWord;
                      if (ww = 0) or (ww > 128) then
                        TxChar(NAK);
                      else
                        if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                          if F16_BlockRead(fh, @SectBuff, ww, ww) then
                            Write(TxChar, ACK + IntToHex(ww));
                            TxBlock(ww);
                          else
                            TxChar(NAK);
                          endif;
                        else
                          TxChar(NAK);
                        endif;
                      endif;
                      SendResponse;
                    |
      cmdBlockWrite : // write a block of data into a file
                      TxChar(char(cmdBlockWrite));
                      RxFileHandle;
                      ww:= RxHexWord;
                      if (ww = 0) or (ww > 128) then
                        TxChar(NAK);
                      else
                        RxBlock(ww);
                        if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                          if F16_BlockWrite(fh, @SectBuff, ww, ww) then
                            Write(TxChar, ACK + IntToHex(ww));
                          else
                            TxChar(NAK);
                          endif;
                        else
                          TxChar(NAK);
                        endif;
                      endif;
                      SendResponse;
                    |
      cmdFileClose  : // close a file
                      TxChar(char(cmdFileClose));
                      RxFileHandle;
                      if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                        F16_FileClose(fh);
                        TxChar(ACK);
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileCreate : // create a new file and preset its size
                      TxChar(char(cmdFileCreate));
                      RxFilePath;
                      RxFileName;
                      RxFileAttr;
                      RxFileTime;
                      RxFileDate;
                      lw:= RxHexLW;
                      if (MediaState = msInit) and (CommErr = 0) then
                        if F16_FileCreate(fPath, fName, fAttr, fTime, fDate, lw) then
                          TxChar(ACK);
                        else
                          TxChar(NAK);
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdRandWrite  : // random write a block of data into a file
                      TxChar(char(cmdRandWrite));
                      RxFileHandle;
                      ww:= RxHexWord;
                      if (ww = 0) or (ww > 128) then
                        TxChar(NAK);
                      else
                        RxBlock(ww);
                        if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                          if F16_BlockRandomWrite(fh, @SectBuff, ww, ww) then
                            Write(TxChar, ACK + IntToHex(ww));
                          else
                            TxChar(NAK);
                          endif;
                        else
                          TxChar(NAK);
                        endif;
                      endif;
                      SendResponse;
                    |
      cmdEndOfFile  : // request for file pointer = end of the file
                      TxChar(char(cmdEndOfFile));
                      RxFileHandle;
                      if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                        if F16_EndOfFile(fh) then
                          TxChar(char(true));
                        else
                          TxChar(char(false));
                        endif;
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdFileSizeH  : // get the byte count of an open file
                      TxChar(char(cmdFileSizeH));
                      RxFileHandle;
                      if HandleOk and (MediaState = msInit) and (CommErr = 0) then
                        lw:= F16_FileSizeH(fh);
                        Write(TxChar, ACK + LongToHex(lw));
                      else
                        TxChar(NAK);
                      endif;
                      SendResponse;
                    |
      cmdRawAccess  : //
                      TxChar(char(cmdRawAccess));
                      case RxChar of
                        'R' : // read a physical sector
                              if DiskState = dsNoDisk then
                                TxChar(NAK);
                              elsif not F16_DiskInit then
                                TxChar(NAK);
                              else
                                TxChar(ACK);
                              endif;
                              lw:= RxHexLw;
                              if CommErr = 0 then
                                _BootOffs:= 0;
                                if F16_ReadSector(lw, @SectBuff) then
                                  TxChar(ACK);
                                  pb:= @SectBuff;
                                  ww:= 128;
                                  while ww > 0 do
                                    TxChar(char(pb^++));
                                    dec(ww);
                                  endwhile;
                                  SendResponse;
                                  ww:= 128;
                                  while ww > 0 do
                                    TxChar(char(pb^++));
                                    dec(ww);
                                  endwhile;
                                  SendResponse;
                                  ww:= 128;
                                  while ww > 0 do
                                    TxChar(char(pb^++));
                                    dec(ww);
                                  endwhile;
                                  SendResponse;
                                  ww:= 128;
                                  while ww > 0 do
                                    TxChar(char(pb^++));
                                    dec(ww);
                                  endwhile;
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                        'W' : // write a physical sector
                              if DiskState in[dsNoDisk, dsWriteProt] then
                                TxChar(NAK);
                              elsif not F16_DiskInit then
                                TxChar(NAK);
                              else
                                TxChar(ACK);
                              endif;
                              lw:= RxHexLw;
                              if CommErr = 0 then
                                TxChar(ACK);
                                pb:= @SectBuff;
                                ww:= 128;
                                while ww > 0 do
                                  pb^++:= byte(RxChar);
                                  dec(ww);
                                endwhile;
                                SendResponse;
                                ww:= 128;
                                while ww > 0 do
                                  pb^++:= byte(RxChar);
                                  dec(ww);
                                endwhile;
                                SendResponse;
                                ww:= 128;
                                while ww > 0 do
                                  pb^++:= byte(RxChar);
                                  dec(ww);
                                endwhile;
                                SendResponse;
                                ww:= 128;
                                while ww > 0 do
                                  pb^++:= byte(RxChar);
                                  dec(ww);
                                endwhile;
                                _BootOffs:= 0;
                                if F16_WriteSector(lw, @SectBuff) then
                                  TxChar(ACK);
                                else
                                  TxChar(NAK);
                                endif;
                              else
                                TxChar(NAK);
                              endif;
                            |
                      else
                        TxChar('?');
                      endcase;
                      SendResponse;
                    |
    else
     { TxChar('?');
      SendResponse;}
    endcase;

    // house keepings
    if Not CardDet then                     // 1 = no card
      DiskState:= dsNoDisk;
      MediaState:= msUnKnown;
    elsif DiskState = dsNoDisk then    // new card inserted
      DiskState:= dsDiskOk;
      MediaState:= msDiskChanged;
    endif;
//    DiskState  = (dsNoDisk, dsWriteProt, dsDiskErr, dsDiskOk);
//    MediaState = (msUnKnown, msDiskChanged, msInit);
  endloop;
end MMC_M64I2C.





